
import os
import os.path
import asyncio
from argparse import ArgumentParser
from asyncio import CancelledError
from typing_extensions import Final

from grpclib.server import Server  # type: ignore
from pymap.interfaces.backend import BackendInterface, ServiceInterface

from .handlers import AdminHandlers

__all__ = ['AdminService']


class AdminService(ServiceInterface):
    """A pymap service implemented using a `grpc <https://grpc.io/>`_ server
    to perform admin functions on a running backend.

    """

    def __init__(self, path: str, server: Server) -> None:
        super().__init__()
        self.path: Final = path
        self.server: Final = server

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('admin service')
        group.add_argument('--admin-path', metavar='PATH',
                           help='path to POSIX socket file')
        group.add_argument('--no-filter', action='store_true',
                           help='do not filter appended messages')

    @classmethod
    async def init(cls, backend: BackendInterface) -> 'AdminService':
        config = backend.config
        if config.args.admin_path is not None:
            path = config.args.admin_path
        else:
            pid = os.getpid()
            path = os.path.join(os.sep, 'tmp', 'pymap', f'admin-{pid}.sock')
        cls._create_dir(path)
        handlers = AdminHandlers(backend)
        server = Server([handlers], loop=asyncio.get_event_loop())
        return cls(path, server)

    @classmethod
    def _create_dir(cls, path: str) -> None:
        try:
            dirname = os.path.dirname(path)
            os.mkdir(dirname, 0o755)
        except FileExistsError:
            pass

    def _unlink_path(self) -> None:
        try:
            os.unlink(self.path)
        except OSError:
            pass

    async def run_forever(self) -> None:
        server = self.server
        await server.start(path=self.path)
        try:
            await server.wait_closed()
        except CancelledError:
            server.close()
            await server.wait_closed()
        finally:
            self._unlink_path()



"""IMAP server with pluggable Python backends."""

import asyncio
import logging
from argparse import ArgumentParser, Namespace, ArgumentDefaultsHelpFormatter
from typing import Any, Type, Sequence, Mapping

from pkg_resources import iter_entry_points, DistributionNotFound

from . import __version__
from .interfaces.backend import BackendInterface, ServiceInterface

_Backends = Mapping[str, Type[BackendInterface]]
_Services = Mapping[str, Type[ServiceInterface]]


def _load_entry_points(group: str) \
        -> Mapping[str, Type[Any]]:
    ret = {}
    for entry_point in iter_entry_points(group):
        try:
            cls = entry_point.load()
        except DistributionNotFound:
            pass  # optional dependencies not installed
        else:
            ret[entry_point.name] = cls
    return ret


def main() -> None:
    parser = ArgumentParser(description=__doc__,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('--debug', action='store_true',
                        help='increase printed output for debugging')
    parser.add_argument('--version', action='version',
                        version='%(prog)s' + __version__)
    parser.add_argument('--no-service', dest='skip_services', action='append',
                        metavar='NAME', help='do not run the given service')
    subparsers = parser.add_subparsers(dest='backend',
                                       help='which pymap backend to use')

    backends: _Backends = _load_entry_points('pymap.backend')
    services: _Services = _load_entry_points('pymap.service')

    for backend_cls in backends.values():
        backend_cls.add_subparser(subparsers)
    for service_cls in services.values():
        service_cls.add_arguments(parser)
    parser.set_defaults(skip_services=[])
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.INFO)
        logging.getLogger(__package__).setLevel(logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    if not args.backend:
        parser.error('Expected backend name')
    backend_type = backends[args.backend]

    service_types = [service for name, service in services.items()
                     if name not in args.skip_services]

    try:
        return asyncio.run(run(args, backend_type, service_types), debug=False)
    except KeyboardInterrupt:
        pass


async def run(args: Namespace, backend_type: Type[BackendInterface],
              service_types: Sequence[Type[ServiceInterface]]) -> None:
    backend = await backend_type.init(args)
    backend.config.apply_context()
    services = [await service.init(backend) for service in service_types]

    await asyncio.gather(*(service.run_forever() for service in services))



"""Admin functions for a running pymap server."""

import os
import os.path
import re
import asyncio
from argparse import ArgumentParser, Namespace

from grpclib.client import Channel  # type: ignore
from pymap import __version__

from .append import AppendCommand
from .command import ClientCommand
from ..grpc.admin_grpc import AdminStub


def _find_path(parser: ArgumentParser) -> str:
    dirname = os.path.join(os.sep, 'tmp', 'pymap')
    try:
        paths = [os.path.join(dirname, fn)
                 for fn in os.listdir(dirname)
                 if re.match(r'^admin-\d+\.sock$', fn)]
    except FileNotFoundError:
        paths = []
    if len(paths) == 1:
        return paths[0]
    parser.error('Cannot determine admin socket path')


def main() -> None:
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('--version', action='version',
                        version='%(prog)s' + __version__)
    parser.add_argument('--path', help='path to admin socket file')

    subparsers = parser.add_subparsers(dest='command',
                                       help='which admin command to run')
    commands = dict([AppendCommand.init(parser, subparsers)])
    args = parser.parse_args()

    if not args.command:
        parser.error('Expected command name.')
    command = commands[args.command]

    asyncio.run(run(parser, args, command), debug=False)


async def run(parser: ArgumentParser, args: Namespace,
              command: ClientCommand) -> None:
    loop = asyncio.get_event_loop()
    path = args.path or _find_path(parser)
    channel = Channel(path=path, loop=loop)
    stub = AdminStub(channel)
    try:
        ret = await command.run(stub, args)
        print(ret)
    finally:
        channel.close()
